//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%

// Background
#include <XSModel/Data/BackCorr/Background.h>
// SpectralData
#include <XSModel/Data/SpectralData.h>
// Weight
#include <XSStat/Weight.h>
// Model
#include <XSModel/Model/Model.h>
// ChiSquare
#include <XSStat/ChiSquare.h>

#include <XSModel/Data/DataUtility.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/GlobalContainer/DataSetTypes.h>
#include <XSModel/Parameter/ModParam.h>
#include <XSModel/Data/Detector/Response.h>
#include <XSUtil/Utils/XSstream.h>
#include <XSUtil/Utils/XSutility.h>
#include <XSUtil/Numerics/Gamma.h>
#include <XSUtil/Error/Error.h>
#include <XSstreams.h>
#include <XSContainer.h>
#include <XSsymbol.h>
#include <iostream>
#include <iomanip>
#include <typeinfo>
#include <set>
#include <numeric>
#include <sstream>
#include <cmath>


static string s_longName = "Chi-Squared";
static string s_greekName = "\\gx\\u2\\d";

// Class ChiSquare 

ChiSquare::ChiSquare()
  : FitStatMethod("chi"), 
    m_hasZeroVarChans(false),
    m_difference(),
    m_variance(),
    m_dSigma(),
    m_weightMethod(0)
{
   canUseChainRule(true);
}

ChiSquare::~ChiSquare()
{
   if (m_weightMethod)
      delete m_weightMethod;
}

void ChiSquare::doPerform (const SpectralData* sp)
{

  // alright, now we perform the sum of the statistic over the models.
  Real chiSquare(0.);
  const size_t specNum = sp->spectrumNumber();
  const std::valarray<size_t>& IN = sp->indirectNotice();

  doReset(sp);

  // Check and fix zero-variance bins PRIOR to adding any model
  // systematic variance. But if "model" weighting is in use,
  // we can only check AFTER weighting is applied.
  m_hasZeroVarChans = false;

  if (m_weightMethod->name() != string("model")) {
    RealArray& var  = m_variance;
    if (applyMinVariance(sp, var)) {
      m_hasZeroVarChans = true;
    }
  }

  XSContainer::ModelMapConstIter ml (XSContainer::models->modelSet().begin());      
  XSContainer::ModelMapConstIter  mlEnd (XSContainer::models->modelSet().end());      
  while ( ml != mlEnd )
  {
     // D - F for each model. In the case where there are multiple
     // sources difference will be iteratively defined.
     Model* m (ml->second);
     if (m->isActive())
     {
       ArrayContainer::const_iterator itFluxEnd = m->modelFlux().end();
       ArrayContainer::const_iterator itFlux = m->modelFlux().find(specNum);
       if ( itFlux != itFluxEnd ) {
	 if ( m_difference.size() != IN.size() ) {
	   std::ostringstream oss;
	   oss << "Array size for spectrum " << specNum << " in ChiSquare::doPerform.";
	   throw RedAlert(oss.str());
	 }
	 m_difference -= m->foldedModel(specNum)[IN];
	 if ( tpout.maxChatter() >= 33 ) {
	   RealArray spec (sp->spectrum()[IN]);
           tcout << " difference " << specNum << " source " << m->sourceNumber() << '\n';
           size_t N (m_difference.size());
           for (size_t j = 0; j < N; ++j) {
	     tcout  << j  << " -= " << m_difference[j] << "  " 
		    << "  " << spec[j] << "  " <<  m->foldedModel(specNum)[IN[j]] << '\n';
           }
           if (N)
              tcout << " spcounts " << spec.sum() << " mcounts " << m->foldedModel(specNum)[IN].sum() << '\n';
           tcout << std::flush;
	 }
       }
       m_weightMethod->addModelVariance(*m,sp,m_variance);
     }
     ++ml;
  }

  // Now do check for case of model weighting.
  if (m_weightMethod->name() == string("model")) {
    const RealArray& var = m_variance;
    const size_t varSz = var.size();
    for (size_t i=0; i<varSz; ++i) {
      if (var[i] == 0.0) {
	std::ostringstream oss;
	oss << "Cannot calculate chi-square using model weighting on spectrum "<< specNum 
	    <<"\nModel weighting causes it to contain 1 or more channels with zero variance.\n\n";
	value(-1.0);
	throw YellowAlert(oss.str());
      } 
    }
  }    

  // Now calculate chi-square    
  const RealArray& diff = m_difference;
  RealArray& var  = m_variance;
  RealArray quot  (diff*diff/var);
  if (quot.size()) {
    chiSquare += quot.sum();
  }

  tpout << xsverbose(35);
  if ( tpout.maxChatter() >= 35 ) {
    tcout << "ChiSquare::doPerform output" <<std::endl;
    tcout << " i data  diff  var \n" ;
    size_t N (diff.size());
    std::vector<Real> PS(N);
    std::partial_sum(&quot[0],&quot[0]+N,PS.begin());

    for (size_t i = 0; i < N; ++i) {
      tcout << i << "  " << sp->spectrum()[i] 
	    << " " << diff[i]
	    << " " << var[i] << " "  << quot[i]
	    << " " << PS[i]  << '\n';
    }
    tcout << std::endl;
  } 
  tpout << xsverbose();

  value(chiSquare);
}

bool ChiSquare::checkWeight (const Weight* weightMethod) const
{
  // ChiSquare is compatible with all of the current Weight methods.
  return true;
}

void ChiSquare::doInitialize (const SpectralData* sp)
{

  // Initialize the weighting method based on the modifier set in StatMethod
  string method = modifier();
  if ( method == "" ) method = "standard";
  if (m_weightMethod)
     delete m_weightMethod;
  m_weightMethod = Weight::createWeightObject(method);


  // deal with the common operations: initialize arrays needed for holding
  // derivatives (if required by fit's FitMethod).

  FitStatMethod::doInitialize(sp);   

  // now, initialize arrays that are specific to ChiSquare.
  // each is an array of arrays of size # of noticed channels,
  // one per spectrum loaded. 

  // for every weighting technique except model weighting, the variance
  // is constant (independent of the model) apart from a possible systematic
  // term we can add later. If the weighting is Model, then the data variance
  // is given by the sum of the model weight and the background variance. 
  // so here we need to store the background variance in the variance array.
  // If model systematic errors or model errors are called for these will be
  // added during iteration by the Weight::addModelVariance() function.


  // protoWorkspace has been set up by FitStatMethod::doInitialize
  // so pw->first is the spectrum number an pw->second an array of size
  // the number of noticed channels.

  m_variance.resize(protoWorkspace().size());
  m_difference.resize(protoWorkspace().size());

  // load the m_variance array with the data variance which will
  // not change during the fit

  m_weightMethod->calcDataVariance(sp, m_variance);
  
  // doReset loads the m_difference array and is done every step of
  // the fit
  doReset(sp);
}

void ChiSquare::doDeallocate ()
{
  FitStatMethod::doDeallocate();      

  m_variance.resize(0);
  m_difference.resize(0);
  m_dSigma.resize(0);
}

void ChiSquare::doReset (const SpectralData* sp)
{

  // reset quantities specific to ChiSquare object for new iteration.

  const std::valarray<size_t>& IN = sp->indirectNotice();
  // Subtle note: Because we are retrieving a subset from
  // a non-const valarray, if we don't do the static cast
  // it will remain as type indirect_array<Real>.  For some
  // reason on Solaris, this then causes a memory leak when 
  // assigned to m_difference.  The cast seems to get
  // rid of this.  
  m_difference = RealArray(sp->spectrum()[IN]);
  if (sp->background()) {
    m_difference -= sp->background()->spectrum()[IN];
  }
  if (sp->correction() && sp->correctionScale() != 0) {
    m_difference -=  sp->correctionScale()
                        *sp->correction()->spectrum()[IN];         
  }     
  m_weightMethod->calcDataVariance(sp, m_variance);

  m_dSigma = 0.0;

}

void ChiSquare::doReportWarnings (const SpectralData* sp) const
{
  using namespace std;
  if (m_hasZeroVarChans) {
    tcout <<"\n***Warning: Chi-square may not be valid due to bins with zero variance"
	  <<"\n            in spectrum number: " << sp->spectrumNumber() << std::endl;
  }
}

Real ChiSquare::statValueForSingleBin (Real obs, Real mod, Real error, Real areaTime, const Real& NODATA) const
{
  Real point (obs - mod > 0 ? (obs - mod)*(obs - mod) : -(obs - mod)*(obs - mod) );
  Real msyst (XSContainer::models->modelSystematicError());
  Real var   (msyst > 0 ? error*error + msyst*msyst*mod*mod : error*error);
  point /= var;

  return point;       
}

Real ChiSquare::deltaSigmaForSingleBin (Real obs, Real mod, Real error, Real areaTime, const Real& NODATA) const
{
  Real point ( obs - mod );
  Real msyst (XSContainer::models->modelSystematicError());
  Real var ( msyst > 0 ?  sqrt(error*error + msyst*msyst*mod*mod) : error ); 
  point /= var;
  return point;
}

RealArray ChiSquare::getDifferences (const SpectralData* sp) const
{
  // seems a waste perhaps to return just a copy of previously calculated array,
  // but not all statistics will have calculated this.
  // the object of providing this virtual function is to support
  // peak residual calculation - may be there are other uses also.
  return m_difference;
}

Real ChiSquare::sumSecondDerivs (const std::map<size_t,RealArray>& dFolded, int pi, int pj)
{

  Real alpha_ij=.0;
  // Calculate (2/var)*(dF/dpi)(dF/dpj)


  const RealArray& piDeriv = dFolded.find(pi)->second;
  const RealArray& pjDeriv = dFolded.find(pj)->second;

  RealArray secondDeriv(piDeriv);
  secondDeriv *= pjDeriv;
  secondDeriv /= m_variance;

  // there needs to be something which knows whether the correction for a model
  // dependence on the variance needs to be made
  RealArray sigmaCorrection;
  calcD2Sigma(sigmaCorrection);
  secondDeriv *= sigmaCorrection;
  
  if (secondDeriv.size())
     alpha_ij += secondDeriv.sum();
  
  alpha_ij *= 2.0;
  return alpha_ij;
}

Real ChiSquare::sumDerivs (const RealArray& dFolded, const int spectrumNumber, 
			   const int parameterIndex) const
{
    // The m_difference array will contain the quantity
    // d = (D-F), the m_variance array will contain the variance var, 
    // corrected for any model systematic terms. Calculation of m_difference
    // is done in Model::difference(ArrayContainer&) which is invoked by
    // ChiSquare::doPerform

    // The derivative of X^2 wrt each parameter j for each folded model F is then 
    // -2(D-F)(1/var_i +  (D-F)/(sigma_i)^3 [dsigma_i/dF])dF_i/dpj
    // the second term is the 'weight derivative' and is provided by the
    // Weight::varianceDeriv function for each array in the model.

    // dF/dp is independent of StatMethod, and so it is calculated one
    // time in StatManager and passed to StatMethods as an argument.

    // dsigma/dF is also independent of StatMethod, but since Cstat/Lstat
    // do not need the 'weight derivative' at all (due to dStat/dsigma = 0),
    // it is calculated and stored in local getWeightDerivs function.

    RealArray dSdpA (m_difference);        
    dSdpA *= dFolded;
    dSdpA /= m_variance;
    dSdpA *= (1. + m_difference/m_variance*m_dSigma);

    Real dSdpj = 0.0;
    if (dSdpA.size())
       dSdpj = -2.0*dSdpA.sum();

    tpout << xsverbose(33);         
    if ( tpout.maxChatter() >= 33 ) {
      size_t N = m_difference.size();
      std::vector<Real> PS(N);
      if (N)
	 std::partial_sum(&dSdpA[0],&dSdpA[0]+N,PS.begin());
      tcout << " Spectrum " << spectrumNumber << " Parameter# " << parameterIndex  
	    << " D:  " << dSdpj << '\n';
      for (size_t j = 0; j < N; ++ j) {
	tcout << j << " " << std::setw(12) << m_difference[j]  << " "
	      << std::setw(12) << m_variance[j] << " "
	      << std::setw(12) << dFolded[j]  << " "
	      << std::setw(12) << dSdpA[j] << " "
	      << std::setw(12) << PS[j]  << '\n'; 
      }
      tcout << " d " << spectrumNumber << " " << dSdpj << '\n';  
    }
    tpout << xsverbose();
    
    return dSdpj;
}

void ChiSquare::calcD2Sigma (RealArray& correction) const
{
  // If variance has a model dependency, calculate contribution
  // to second derivatives.
  // Let g = ((D-F)*sigma*d(sigma)/dF)/sigma^2,  then correction
  // = (1 + 4g + 3g^2).
    RealArray g(m_dSigma);
    g *= m_difference;
    g /= m_variance;
    correction.resize(g.size());
    correction = g;
    correction *= g;
    correction *= 3.0;
    correction += 4.0*g;
    correction += 1.0;
}

void ChiSquare::getWeightDerivs (const SpectralData* spectrum)
{
  // Remember this is for calculating dsigma/dF, which is 
  // independent of the StatMethods in use.  It actually fills the 
  // m_dSigma arrays with the quantity SIGMA*dsigma/dF, which makes
  // the calculations a little easier here and in sumDerivs.

  XSContainer::ModelMapConstIter mm (XSContainer::models->modelSet().begin());
  XSContainer::ModelMapConstIter mmEnd (XSContainer::models->modelSet().end());
  while ( mm != mmEnd ) {
    Model& m (*mm->second);
    if ( m.isActive()) 
      m_weightMethod->varianceDeriv(m,spectrum,m_dSigma);
    ++mm;       
  }
}

std::pair<Real,Real> ChiSquare::adjustForFrozen (const SpectralData* spec, const std::vector<Model*>& models, const std::map<Model*,ArrayContainer >& savedValsForFrozen, const int nFitSpecs)
{
    // NOTE:  This ASSUMES m_difference and m_variance have been reset to their
    // initial values in doReset, prior to their usage here.  It is perhaps less
    // dangerous to simply call doReset from here, but that would result in many
    // redundant calls (see how adjustForFrozen is called from StatManager::renormalizeStats).
   const std::valarray<size_t>& IN = spec->indirectNotice();
   const size_t N = IN.size(); 
   size_t specNum = spec->spectrumNumber();
   std::pair<Real,Real> fSums(0.,0.);
   if (N)
   {
      const RealArray& pha = m_difference;
      RealArray totalModel(0.,N);
      RealArray totModFroz(.0,N);
      RealArray var(m_variance);
      applyMinVariance(spec, var);
      for (size_t i=0; i<models.size(); ++i)
      {
         Model* mod = models[i];
         std::map<Model*,ArrayContainer >::const_iterator itSaved =
                   savedValsForFrozen.find(mod);
         if (itSaved != savedValsForFrozen.end())
         {
            totalModel += itSaved->second.find(specNum)->second[IN];
            totModFroz += mod->foldedModel(specNum)[IN];
         }
         else
         {
            totalModel += mod->foldedModel(specNum)[IN];
         }
      }
      RealArray totModNonFroz(totalModel - totModFroz);
      RealArray obsNonFroz((pha - totModFroz)*totModNonFroz);
      fSums.first = (obsNonFroz/var).sum();
      fSums.second = (totModNonFroz*totModNonFroz/var).sum();

      //print frozen and total arrays to log file for debugging purposes.
      tpout << xsverbose(33);         
      if ( tpout.maxChatter() >= 33 )
      {
         using std::setw;
         tcout.precision(6);
         for ( size_t j = 0; j < N ; ++j)
         {
            tcout   <<  " i " << setw(5) << j 
                    << " obs " << setw(12) << pha[j]
                    << " mod " << setw(12) << totalModel[j] ;
               tcout << " fro " << setw(12) << totModFroz[j];
            tcout << '\n';
         }
         tcout << " ofSum " << fSums.first << " mfSum " << fSums.second 
                   << std::endl;
       }        
      tpout << xsverbose();
   }
   return fSums;
}

bool ChiSquare::applyMinVariance (const SpectralData* sd, RealArray& variance)
{
     bool isRequired = false;
     const std::valarray<size_t>& IN = sd->indirectNotice();
     Real time = sd->exposureTime();
     const size_t sz = variance.size();
     Real backTime = -1.0;
     RealArray bscaleRatio(1.0, IN.size());
     if (sd->background())
     {
        const SpectralData* bckSpec = sd->background()->data();
        backTime = bckSpec->exposureTime();
        bscaleRatio = sd->backgroundScale()[IN]/bckSpec->backgroundScale()[IN];
	bscaleRatio *= sd->areaScale()[IN]/bckSpec->areaScale()[IN];
     }
     for (size_t i=0; i<sz; ++i)
     {
        if (variance[i] == 0.0)
        {
           variance[i] = ChiSquare::calcMinVariance(time, backTime,
                                bscaleRatio[i]);
           isRequired = true;
        }
     }
     return isRequired;
}

Real ChiSquare::calcMinVariance (Real areaTime, Real backAreaTime, Real bscaleRatio)
{
   Real minVar = .0;
   // Use negative backAreaTime as a flag indicating no background file.
   if (backAreaTime < .0)
   {
      minVar = 1.0/(areaTime*areaTime);
   }
   else
   {
      Real factor = 1.0/(areaTime*areaTime);
      Real backFactor = (bscaleRatio*bscaleRatio)/(backAreaTime*backAreaTime);
      minVar = std::min(factor, backFactor);
   }
   return minVar;
}

const std::string ChiSquare::fullName () const
{
  string outstr(s_longName);
  if ( modifier() != "" && modifier() != "standard" ) outstr += "(" + modifier() + ")";
  return outstr;
}

const std::string ChiSquare::scriptName () const
{
  string outstr(s_greekName);
  if ( modifier() != "" && modifier() != "standard" ) outstr += "(" + modifier() + ")";
  return outstr;
}

const std::string ChiSquare::commandName () const
{
  string outstr(name());
  if ( modifier() != "" && modifier() != "standard" ) outstr += " " + modifier();
  return outstr;
}  




// Additional Declarations
